from playhouse.shortcuts import model_to_dict

from pyplanet.apps.contrib.rankings import RankedMap
from pyplanet.apps.core.maniaplanet.models import Map
from pyplanet.views.generics import ManualListView


class TopRanksView(ManualListView):
	title = 'Top ranked players on the server'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, player, top_ranks):
		"""
		Init topsums list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.contrib.rankings.Rankings
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.provide_search = False
		self.top_ranks = top_ranks

	async def get_data(self):
		data = list()
		for idx, rank in enumerate(self.top_ranks):
			data.append(dict(
				rank=idx+1,
				player_nickname=rank.player.nickname,
				average='{:0.2f}'.format((rank.average / 10000))
			))

		return data

	async def destroy(self):
		self.top_ranks = None
		await super().destroy()

	def destroy_sync(self):
		self.top_ranks = None
		super().destroy_sync()

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'rank',
				'sorting': False,
				'searching': False,
				'width': 10,
				'type': 'label'
			},
			{
				'name': 'Player',
				'index': 'player_nickname',
				'sorting': False,
				'searching': False,
				'width': 80,
				'type': 'label'
			},
			{
				'name': 'Average',
				'index': 'average',
				'sorting': False,
				'searching': False,
				'width': 25,
				'type': 'label'
			}
		]


class MapListView(ManualListView):
	model = RankedMap
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Browse'

	def __init__(self, app, player, maps, title, show_rank):
		"""
		Init no-rank list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.contrib.rankings.Rankings
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.maps = maps
		self.title = title
		self.show_rank = show_rank

	async def get_fields(self):
		columns = [
			{
				'name': 'Name',
				'index': 'name',
				'sorting': True,
				'searching': True,
				'search_strip_styles': True,
				'width': 100,
				'type': 'label',
				'action': self.action_jukebox if ('jukebox' in self.app.instance.apps.apps) else None
			},
			{
				'name': 'Author',
				'index': 'author_login',
				'sorting': True,
				'searching': True,
				'search_strip_styles': True,
				'renderer': lambda row, field:
				row['author_nickname']
				if 'author_nickname' in row and row['author_nickname'] and len(row['author_nickname'])
				else row['author_login'],
				'width': 60,
			},
		]

		if self.show_rank:
			columns.append({
				'name': 'Rank',
				'index': 'player_rank',
				'sorting': False,
				'searching': False,
				'search_strip_styles': False,
				'width': 15,
			})

		return columns

	async def get_data(self):
		data = list()

		for list_map in self.maps:
			data.append(model_to_dict(list_map))

		return data

	async def get_actions(self):
		custom_actions = list()
		if 'jukebox' in self.app.instance.apps.apps:
			custom_actions.append(dict(
				name='Add to folder',
				action=self.app.instance.apps.apps['jukebox'].add_to_folder,
				text='&#xf07b;',
				textsize='1.2',
				safe=True,
				type='label',
				order=-50,
				require_confirm=False,
			))

		return custom_actions

	async def action_jukebox(self, player, values, map_info, **kwargs):
		await self.app.instance.apps.apps['jukebox'].add_to_jukebox(player, await self.app.instance.map_manager.get_map(map_info['uid']))



class Queries:
	CALCULATE_WITH_PARTITION = """
-- Reset the current ranks to insert new ones later one.
TRUNCATE TABLE rankings_rank;
-- Limit on maximum ranked records.
SET @ranked_record_limit = {};
-- Minimum amount of ranked records required to acquire a rank.
SET @minimum_ranked_records = {};
-- Total amount of maps active on the server.
SET @active_map_count = {};
-- Set the rank/current rank variables to ensure correct first calculation
INSERT INTO rankings_rank (player_id, average, calculated_at)
SELECT
	player_id, average, calculated_at
FROM (
	SELECT
		player_id,
		-- Calculation: the sum of the record ranks is combined with the ranked record limit times the amount of unranked maps.
		-- Divide this summed ranking by the amount of active maps on the server, and an average calculated rank will be returned.
		ROUND((SUM(player_rank) + (@active_map_count - COUNT(player_rank)) * @ranked_record_limit) / @active_map_count * 10000, 0) AS average,
		NOW() AS calculated_at,
		COUNT(player_rank) AS ranked_records_count
	FROM
	(
		SELECT
			id,
			map_id,
			player_id,
			score,
			RANK() OVER (PARTITION BY map_id ORDER BY score ASC) AS player_rank
		FROM localrecord
		WHERE map_id IN ({})
	) AS ranked_records
	WHERE player_rank <= @ranked_record_limit
	GROUP BY player_id
) grouped_ranks
WHERE ranked_records_count >= @minimum_ranked_records
"""

	CALCULATE_WITHOUT_PARTITION = """
-- Reset the current ranks to insert new ones later one.
TRUNCATE TABLE rankings_rank;
-- Limit on maximum ranked records.
SET @ranked_record_limit = {};
-- Minimum amount of ranked records required to acquire a rank.
SET @minimum_ranked_records = {};
-- Total amount of maps active on the server.
SET @active_map_count = {};
-- Set the rank/current rank variables to ensure correct first calculation
SET @player_rank = 0;
SET @current_rank = 0;
INSERT INTO rankings_rank (player_id, average, calculated_at)
SELECT
	player_id, average, calculated_at
FROM (
	SELECT
		player_id,
		-- Calculation: the sum of the record ranks is combined with the ranked record limit times the amount of unranked maps.
		-- Divide this summed ranking by the amount of active maps on the server, and an average calculated rank will be returned.
		ROUND((SUM(player_rank) + (@active_map_count - COUNT(player_rank)) * @ranked_record_limit) / @active_map_count * 10000, 0) AS average,
		NOW() AS calculated_at,
		COUNT(player_rank) AS ranked_records_count
	FROM
	(
		SELECT
			id,
			map_id,
			player_id,
			score,
			@player_rank := IF(@current_rank = map_id, @player_rank + 1, 1) AS player_rank,
			@current_rank := map_id
		FROM localrecord
		WHERE map_id IN ({})
		ORDER BY map_id, score ASC
	) AS ranked_records
	WHERE player_rank <= @ranked_record_limit
	GROUP BY player_id
) grouped_ranks
WHERE ranked_records_count >= @minimum_ranked_records
"""



import random

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.funcmd.view import EmojiToolbarView
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting

from pyplanet.utils.style import style_strip


class FunCmd(AppConfig):
	app_dependencies = ['core.maniaplanet']

	muffins = [
		'Muffin',
		'Apple Cinnamon Muffin',
		'Cornbread Muffin',
		'Blueberry Cream Cheese Muffin',
		'Sweet Potato Muffin',
		'Chocolate Muffin',
		'Coffee Cake Muffin',
		'Snickerdoodle Mini Muffin',
		'Oatmeal Muffin',
		'Cinnamon Streusel Muffin',
		'Apple Cider Muffin',
		'Toffe-ee Muffin ;-)',
		'Insert-Developer-Joke-Here Muffin',
		'Orange Marmalade Muffin',
		'Pumpkin Spice Muffin',
		'Cranberry Oatmeal Muffin',
		'Pineapple & Sour Cream Muffin',
		'Lemon Yogurt Muffin',
		'Zucchini Muffin',
		'Slice of Apple Pie',
		'Baklava',
		'Bowl of Gelato',
		'Picarones',
		'Syrniki',
		'Lamingtons',
		'Cup of Skyr',
		'Om Ali',
		'Bread Pudding',
		'Loukoumades',
		'Slice of Spekkoek',
		'Slice of Slagroomtaart',
		'Appelflappen',
		'Banana Muffin',
		'No Muffin',
		'Blueberry Muffin',
		'Two Muffins',
		'Muffin with Cream',
		'Choc-Chip Muffin',
		'Double Chocolate Muffin',
		'Lemmon & Poppyseed Muffin',
		'Apple Crunch Muffin',
		'Raspberry Muffin',
		'Pink Cupcake',
		'Pumpkin Pie',
		'Slice of Cheesecake',
		'Cinnamon Doghnut',
		'Carrot Cake',
	]

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.setting_emoji_toolbar = Setting(
			'emoji_toolbar', 'Display Emoji Toolbar', Setting.CAT_DESIGN, type=bool, default=True,
			description='Display the Emoji Toolbar to users.',
			change_target=self.reload_settings
		)

		self.emoji_toolbar = EmojiToolbarView(self)

	async def on_start(self):
		await self.context.setting.register(
			self.setting_emoji_toolbar
		)

		await self.instance.command_manager.register(
			Command(command='afk', target=self.command_afk, admin=False, description='Set yourself as AFK'),
			Command(command='bootme', target=self.command_bootme, admin=False, description='Boot yourself from the server'),
			Command(command='rq', aliases=['ragequit'], target=self.command_rq, admin=False, description='Ragequit from the server'),
			Command(command='gg', target=self.command_gg, admin=False, description='Send Good Game to everyone'),
			Command(command='n1', target=self.command_n1, admin=False, description='Send Nice One to everyone'),
			Command(command='nt', target=self.command_nt, admin=False, description='Send Nice Try/Nice Time to everyone'),
			Command(command='muffin', target=self.command_muffin, admin=False, description='Give muffin to another player').add_param(name='login', required=True),
		)

		if self.instance.game.game == 'sm':
			await self.instance.command_manager.register(
				Command(command='ns', target=self.command_ns, admin=False, description='Send Nice Shot to everyone'),
			)

		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		if await self.setting_emoji_toolbar.get_value():
			await self.emoji_toolbar.display()

	async def reload_settings(self, *args, **kwargs):
		if await self.setting_emoji_toolbar.get_value():
			await self.emoji_toolbar.display()
		else:
			await self.emoji_toolbar.hide()

	async def player_connect(self, player, *args, **kwargs):
		if await self.setting_emoji_toolbar.get_value():
			await self.emoji_toolbar.display(player_logins=[player.login])

	async def command_afk(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.gbx.multicall(
			self.instance.gbx('ForceSpectator', player.login, 3),
			self.instance.chat('$fff {}$z$s$fff is now away from keyboard.'.format(player.nickname))
		)

	async def command_muffin(self, player, data, **kwargs):
		# Try to match the player.
		if self.instance.game.game == 'tmnext':
			nickname_input = style_strip(data.login).lower()
			dest_player = [p for p in self.instance.player_manager.online if style_strip(p.nickname.lower()) == nickname_input]
		else:
			dest_player = [p for p in self.instance.player_manager.online if p.login == data.login]

		# Try to get the nickname of the destination player, otherwise just show the input login.
		if dest_player:
			dest_nickname = dest_player[0].nickname
		else:
			dest_nickname = data.login

		# Prepare the message.
		muffin = random.choice(self.muffins)
		a_word = 'a'
		if muffin[0:1].lower() in ['a', 'e', 'u', 'i', 'o']:
			a_word = 'an'

		message = '$z$s{}$z$s$m$f90 gave {} {} to $z$s{}'.format(
			player.nickname, a_word, muffin, dest_nickname
		)

		# Send the message.
		await self.instance.chat(message)

	async def command_bootme(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.gbx.multicall(
			self.instance.chat('$fff  {}$z$s$fff chooses to boot back to real life!'.format(player.nickname)),
			self.instance.gbx('Kick', player.login, 'chooses to boot to real life (/bootme)'),
		)

	async def command_rq(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.gbx.multicall(
			self.instance.chat('$f00 {}$z$s$f00 rage quits.'.format(player.nickname)),
			self.instance.gbx('Kick', player.login, 'rage quit (/rq)'),
		)

	async def command_gg(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.chat('$fff {}$z$s$fff Good Game everyone!'.format(player.nickname))

	async def command_n1(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.chat('$fff {}$z$s$fff Nice one!'.format(player.nickname))

	async def command_ns(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		await self.instance.chat('$fff {}$z$s$fff Nice shot!'.format(player.nickname))

	async def command_nt(self, player, data, **kwargs):
		if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
			return
		if self.instance.game.game == 'sm':
			await self.instance.chat('$fff {}$z$s$fff Nice try!'.format(player.nickname))
		else:
			await self.instance.chat('$fff {}$z$s$fff Nice time!'.format(player.nickname))



import logging
import math
from packaging import version
from peewee import RawQuery

from pyplanet.apps.contrib.rankings.models.ranked_map import RankedMap
from pyplanet.apps.contrib.rankings.models import Rank
from pyplanet.apps.contrib.rankings.queries import Queries
from pyplanet.apps.contrib.rankings.views import TopRanksView, MapListView
from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting

logger = logging.getLogger(__name__)


class Rankings(AppConfig):
	# Rankings can only be calculated on Trackmania games.
	game_dependencies = ['trackmania', 'trackmania_next']

	# Rankings depend on the local records.
	app_dependencies = ['core.maniaplanet', 'core.trackmania', 'local_records']

	# Whether the system supports using the partition query.
	supports_partition = False

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.setting_records_required = Setting(
			'minimum_records_required', 'Minimum records to acquire ranking', Setting.CAT_BEHAVIOUR, type=int,
			description='Minimum of records required to acquire a rank (minimum 3 records).',
			default=5
		)

		self.setting_chat_announce = Setting(
			'rank_chat_announce', 'Display server ranks on map start', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether to display the server rank on every map start.',
			default=True
		)

		self.setting_topranks_limit = Setting(
			'topranks_limit', 'Maximum rank to display in topranks', Setting.CAT_BEHAVIOUR, type=int,
			description='Amount of ranks to display in the topranks view.',
			default=100
		)

	async def on_start(self):
		await self.check_database_compatibility()

		# Listen to signals.
		self.context.signals.listen(mp_signals.map.map_end, self.map_end)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		# Register commands.
		await self.instance.command_manager.register(
			Command('rank', target=self.chat_rank, description='Displays your current server rank.'),
			Command('nextrank', target=self.chat_nextrank, description='Displays the player ahead of you in the server ranking.'),
			Command('topranks', target=self.chat_topranks, description='Displays a list of top ranked players.'),
		)

		# Register settings
		await self.context.setting.register(self.setting_records_required, self.setting_chat_announce, self.setting_topranks_limit)

	async def check_database_compatibility(self):
		# The queries used for calculating the rankings are only compatible with MySQL and MariaDB.
		# Provide an error when running PostgreSQL.
		if self.instance.db.server_info.type == "postgresql":
			raise NotImplementedError("Rankings app only works on PyPlanet instances running on MySQL.")

		logger.info('DB Information: {} ({})'.format(self.instance.db.server_info.type, self.instance.db.server_info.version))
		# Database engines starting from MySQL 8.0 / MariaDB 10.2 support the PARTITION BY query.
		# The query without PARTITION BY is unsupported in newer versions.
		if self.instance.db.server_info.type == "mysql" and \
			version.parse(self.instance.db.server_info.version) >= version.parse("8.0"):
			self.supports_partition = True
		elif self.instance.db.server_info.type == "mariadb" and \
			version.parse(self.instance.db.server_info.version) >= version.parse("10.2"):
			self.supports_partition = True

	async def map_end(self, map):
		# Calculate server ranks.
		await self.calculate_server_ranks()

		# Display the server rank for all players on the server after calculation, if enabled.
		if await self.setting_chat_announce.get_value():
			for player in self.instance.player_manager.online:
				await self.chat_rank(player)

	async def player_connect(self, player, is_spectator, source, signal):
		if await self.setting_chat_announce.get_value():
			await self.chat_rank(player)

	async def calculate_server_ranks(self):
		maps_on_server = [map_on_server.id for map_on_server in self.instance.map_manager.maps]

		minimum_records_required_setting = await self.setting_records_required.get_value()
		minimum_records_required = minimum_records_required_setting if minimum_records_required_setting >= 3 else 3

		maximum_record_rank = await self.get_maximum_record_rank()

		query_text = Queries.CALCULATE_WITH_PARTITION if self.supports_partition else Queries.CALCULATE_WITHOUT_PARTITION

		query = RawQuery(Rank, query_text.format(
			maximum_record_rank, minimum_records_required, str(len(maps_on_server)), ", ".join(str(map_id) for map_id in maps_on_server))
		)

		await Rank.execute(query)

	async def chat_topranks(self, player, *args, **kwargs):
		top_ranks_limit = await self.setting_topranks_limit.get_value()
		top_ranks = await Rank.execute(Rank.select(Rank, Player).join(Player).order_by(Rank.average.asc()).limit(top_ranks_limit))
		view = TopRanksView(self, player, top_ranks)
		await view.display(player)

	async def chat_rank(self, player, *args, **kwargs):
		player_rank = await self.get_player_rank(player)
		if player_rank is None:
			await self.instance.chat('$f00$iYou do not have a server rank yet!', player)
			return

		await self.instance.chat('$f80Your server rank is $fff{}$f80 of $fff{}$f80, average: $fff{}$f80'.format(
			player_rank['rank'], player_rank['total_ranked_players'], player_rank['average']), player)

	async def chat_nextrank(self, player, *args, **kwargs):
		player_ranks = await Rank.execute(Rank.select().where(Rank.player == player.get_id()))

		if len(player_ranks) == 0:
			await self.instance.chat('$f00$iYou do not have a server rank yet!', player)
			return

		player_rank = player_ranks[0]
		next_ranked_players = await Rank.execute(
			Rank.select(Rank, Player)
				.join(Player)
				.where(Rank.average < player_rank.average)
				.order_by(Rank.average.desc())
				.limit(1))

		if len(next_ranked_players) == 0:
			await self.instance.chat('$f00$iThere is no better ranked player than you!', player)
			return

		next_ranked = next_ranked_players[0]
		next_player_rank_average = '{:0.2f}'.format((next_ranked.average / 10000))
		next_player_rank_index = (await Rank.objects.count(Rank.select(Rank).where(Rank.average < next_ranked.average)) + 1)
		next_player_rank_difference = math.ceil((player_rank.average - next_ranked.average) / 10000 * len(self.instance.map_manager.maps))

		await self.instance.chat('$f80The next ranked player is $<$fff{}$>$f80 ($fff{}$f80), average: $fff{}$f80 [$fff-{} $f80RP]'.format(
			next_ranked.player.nickname, next_player_rank_index, next_player_rank_average, next_player_rank_difference), player)

	async def chat_norank(self, player, *args, **kwargs):
		ranked_maps = await self.get_player_map_ranks(player)
		non_ranked_maps = [map for map in self.instance.map_manager.maps if
							map.id not in [ranked_map.id for ranked_map in ranked_maps]]

		view = MapListView(self, player, maps=non_ranked_maps, title='Your non-ranked maps on this server', show_rank=False)
		await view.display(player)

	async def chat_bestrank(self, player, *args, **kwargs):
		ranked_maps = await self.get_player_map_ranks(player, 'ORDER BY player_rank ASC')
		view = MapListView(self, player, maps=ranked_maps, title='Your best ranked maps on this server', show_rank=True)
		await view.display(player)

	async def chat_worstrank(self, player, *args, **kwargs):
		ranked_maps = await self.get_player_map_ranks(player, 'ORDER BY player_rank DESC')
		view = MapListView(self, player, maps=ranked_maps, title='Your worst ranked maps on this server', show_rank=True)
		await view.display(player)

	async def get_player_map_ranks(self, player, sort_query = ''):
		maximum_record_rank = await self.get_maximum_record_rank()

		query = '''SELECT
					map.id,
					map.name,
					map.uid,
					map.author_login,
					ranked_records.player_rank
				FROM
				(
					SELECT
						id,
						map_id,
						player_id,
						score,
						@player_rank := IF(@current_rank = map_id, @player_rank + 1, 1) AS player_rank,
						@current_rank := map_id
					FROM localrecord r,
						(SELECT @player_rank := 0) pr,
						(SELECT @current_rank := 0) cr
					ORDER BY map_id, score ASC
				) AS ranked_records

				INNER JOIN map
				ON map.id = map_id

				WHERE player_rank <= {}
				AND player_id = {} {}'''.format(maximum_record_rank, player.id, sort_query)

		select_query = RawQuery(RankedMap, query)
		ranked_maps = [map for map in await RankedMap.execute(select_query) if
					   map.id in [server_map.id for server_map in self.instance.map_manager.maps]]

		return ranked_maps

	async def get_player_rank(self, player):
		player_ranks = await Rank.execute(Rank.select().where(Rank.player == player.get_id()))

		if len(player_ranks) == 0:
			return None

		player_rank = player_ranks[0]
		player_rank_average = '{:0.2f}'.format((player_rank.average / 10000))
		player_rank_index = (await Rank.objects.count(Rank.select(Rank).where(Rank.average < player_rank.average)) + 1)
		total_ranked_players = await Rank.objects.count(Rank.select(Rank))

		return {'rank': player_rank_index, 'average': player_rank_average, 'total_ranked_players': total_ranked_players}

	async def get_maximum_record_rank(self):
		# Determine the maximum record rank that is included in the locals.
		# The rank calculation requires a non-zero value, so if none is provided it'll default to 1000.
		maximum_record_rank = await self.instance.apps.apps['local_records'].setting_record_limit.get_value()
		if maximum_record_rank == 0:
			maximum_record_rank = 1000

		return maximum_record_rank



import datetime
from peewee import *
from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.core.db import Model


class Rank(Model):
	player = ForeignKeyField(Player, index=True)
	"""
	Player that has the rank.
	"""

	average = IntegerField()
	"""
	Average map ranking.
	"""

	calculated_at = DateTimeField(
		default=datetime.datetime.now,
	)
	"""
	When was the rank calculated?
	"""

	class Meta:
		db_table = 'rankings_rank'



import logging

import peewee_async
from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models.karma import Karma

logger = logging.getLogger(__name__)


def upgrade(migrator: SchemaMigrator):
	# Fix the duplicates in the current setup.
	total_processed = 0
	results = None

	# Make sure that if we are in MySQL, to disable the group only by.
	if isinstance(migrator.database, peewee_async.MySQLDatabase):
		migrator.database.execute_sql("SET sql_mode=(SELECT REPLACE(@@sql_mode,'ONLY_FULL_GROUP_BY',''));")

	while results is None or len(results) > 0:
		results = (Karma.select()
				   .order_by(Karma.created_at.desc())
				   .group_by(Karma.player_id, Karma.map_id)
				   .having(fn.Count(Karma.player_id) > 1))

		for result in results:
			result.delete_instance()
			total_processed += 1

	logger.warning('Removing duplicated karma votes, processed {} rows! Now going to execute the migration...'.format(
		total_processed
	))

	# Create the indexes.
	migrate(
		migrator.add_index(Karma._meta.db_table, [
			'player_id', 'map_id'
		], unique=True),
	)


def downgrade(migrator: SchemaMigrator):
	pass



from pyplanet.utils import style


def test_style_stripping():
	# Colors
	raw = '$i$fffMax$06fSmurf$f00.$fffes$$l$09f.$fffm$08f$a5x$n$w$o'
	expect = '$iMaxSmurf.es$l.mx$n$w$o'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect

	raw = '$l[some link]$i$FFFMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx$l'
	expect = '$l[some link]$iMaxSmurf.esl.mx$l'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect

	raw = '$l[some link]$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx'
	expect = '$l[some link]$iMaxSmurf.esl.mx'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect

	# Links
	raw = '$l$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x$l'
	expect = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	raw = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x'
	expect = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	raw = '$l[some link]$i$fffMax$06fSmurf$f00.$fffes$$l$09f.$fffm$08fx$l'
	expect = '$i$fffMax$06fSmurf$f00.$fffes$l$09f.$fffm$08fx'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	raw = '$l[some link]$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx'
	expect = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	# Sizes
	raw = '$i$n$fffMax$06fSmurf$f00.$w$o$fffe$$nsl$09f.$w$fffm$08f$a5$ox'
	expect = '$fffMax$06fSmurf$f00.$fffe$nsl$09f.$fffm$08f$a5x'
	assert style.style_strip(raw, style.STRIP_SIZES) == expect

	# All
	raw = '$h$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x$h'
	expect = 'MaxSmurf.esl.mx'
	assert style.style_strip(raw) == expect

	raw = '$l[some link]$i$fffMax$06fSmur$$f$f00.$fffesl$09f.$fffm$08fx$l'
	expect = 'MaxSmur$f.esl.mx'
	assert style.style_strip(raw) == expect

	raw = '$l[some link]$i$fffMax$06fSmu$nrf$f00.$fffesl$09f.$fffm$08fx'
	expect = 'MaxSmurf.esl.mx'
	assert style.style_strip(raw) == expect

	# Combination & Misc
	raw = '$f80$i$oToffe$g$z$06fSmurf $z$sHello'
	expect = 'Toffe$zSmurf $zHello'
	assert style.style_strip(raw, keep_reset=True) == expect

	raw = '$f80$i$oToffe$g$z$06fSmurf $z$sHello'
	expect = 'ToffeSmurf Hello'
	assert style.style_strip(raw) == expect

	raw = '$f80$i$oToffe$g$z$06fSmurf $z$sHello'
	expect = 'ToffeSmurf $sHello'
	assert style.style_strip(raw, style.STRIP_SIZES, style.STRIP_COLORS) == expect


def test_bug_1249():
	raw = '$$'
	expect = '$'
	assert style.style_strip(raw) == expect

	raw = '$$ff0Test'
	expect = '$ff0Test'
	assert style.style_strip(raw) == expect

	raw = 'Te$mst'
	expect = 'Test'
	assert style.style_strip(raw) == expect

	raw = 'test mes$9sage'
	expect = 'test message'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect
	assert style.style_strip(raw) == expect




"""
The server information object contains version information about the database server.
"""
import logging
import peewee

logger = logging.getLogger(__name__)


class ServerInfo:
	def __init__(self, engine, db):
		"""
		Initiate database server information object.

		:param engine: Database engine instance.
		:param db: Database instance.
		"""
		self.engine = engine
		self.db = db
		self.type = None
		self.version = None
		self.version_text = None

	async def determine(self):
		if isinstance(self.db.engine, peewee.MySQLDatabase):
			self.type = "mysql"
			query = "SELECT @@version AS version"
		elif isinstance(self.db.engine, peewee.PostgresqlDatabase):
			self.type = "postgresql"
			query = "SHOW SERVER_VERSION"
		else:
			logger.warning("Unable to determine database server version (unknown engine type: {})".format(type(self.db.engine)))
			return

		try:
			cursor = self.engine.execute_sql(query)
			result = cursor.fetchone()

			if result is None or len(result) == 0:
				return

			if self.type == "mysql":
				# Use the innodb_version (1) for a clean engine version, use the version (0) as version text.
				self.version_text = result[0]
				self.version = self.version_text.split('-')[0]

				if "mariadb" in self.version_text.lower():
					self.type = "mariadb"
			elif self.type == "postgresql":
				self.version_text = result[0]
				self.version = self.version_text.split(' ')[0]

			logger.info("Determined database server type: {}, version: {}".format(self.type, self.version))
		except:
			# No database version could be established
			logger.warning("Unable to determine database server version (type: {})".format(self.type))




from peewee import RawQuery, BooleanField
from playhouse.migrate import migrate, SchemaMigrator

from ..models.mapfolder import MapFolder


def upgrade(migrator: SchemaMigrator):
	try:
		query = RawQuery(MapFolder, """SELECT * FROM mapfolder WHERE public = True""")
		public_folders = query.execute()

		migrate(
			migrator.add_column(MapFolder._meta.db_table, 'visibility', MapFolder.visibility),
			migrator.drop_column(MapFolder._meta.db_table, 'public')
		)

		MapFolder.update(visibility='public').where(MapFolder.id << [f.id for f in public_folders]).execute()
	except Exception as e:
		print(str(e))
		print('Migration failed but silencing error!')

def downgrade(migrator: SchemaMigrator):
	try:
		query = RawQuery(MapFolder, """SELECT * FROM mapfolder WHERE visibility = 'public'""")
		public_folders = query.execute()

		migrate(
			migrator.add_column(MapFolder._meta.db_table, 'public', BooleanField()),
			migrator.drop_column(MapFolder._meta.db_table, 'visibility')
		)

		MapFolder.update(public=True).where(MapFolder.id << [f.id for f in public_folders]).execute()
	except Exception as e:
		print(str(e))
		print('Migration failed but silencing error!')



from peewee import *
from pyplanet.core.db import Model


class RankedMap(Model):
	id = IntegerField()
	"""
	ID of the map on which the player has the rank.
	"""

	name = CharField(max_length=150)
	"""
	Name of the map on which the player has the rank.
	"""

	uid = CharField(max_length=50)
	"""
	UID of the map on which the player has the rank.
	"""

	author_login = CharField(max_length=100)
	"""
	Author login of the map on which the player has the rank.
	"""

	player_rank = IntegerField()
	"""
	Rank the player has on the map.
	"""
